#ifndef ATTENDEEINDEX_HPP
#define ATTENDEEINDEX_HPP

#include "Common.hpp"
#include "Checker.hpp"
#include <numeric>

class AttendeeIndex
{
public:
    class Node
    {
    public:
        uint64_t idx;
        double xs[2];
        double ys[2];
        std::unique_ptr<Node> childNodes[2];
    };

    AttendeeIndex(const Problem& problem)
    {
        std::vector<uint64_t> idxs(problem.attendees.size());
        std::iota(idxs.begin(), idxs.end(), (uint64_t)0);
        pRoot = build(problem, idxs);
    }

    template<typename OnBlocked>
    void forEachBlocked(Coord musicianCoord,
                        Coord obstructionCoord,
                        double obstructionRadius,
                        const OnBlocked& onBlocked) const
    {
        AttendeeChecker ac(musicianCoord, obstructionCoord, obstructionRadius);
        forEachBlockedImpl(ac, onBlocked, pRoot.get(), false);
    }

private:
    template<typename OnBlocked>
    void forEachBlockedImpl(const AttendeeChecker& ac,
                            const OnBlocked& onBlocked,
                            const Node* pNode,
                            bool blocked) const
    {
        while (true)
        {
            if (pNode->idx != (uint64_t)-1)
            {
                if (!blocked)
                {
                    blocked = ac.check(Coord(pNode->xs[0], pNode->ys[0]));
                }
                if (blocked)
                {
                    onBlocked(pNode->idx);
                }
                return;
            }

            if (!blocked)
            {
                BlockedStatus blockedStatus =
                    ac.checkRect(pNode->xs, pNode->ys);
                if (blockedStatus == BlockedStatus::Unblocked) return;
                if (blockedStatus == BlockedStatus::Blocked) blocked = true;
            }
            forEachBlockedImpl(ac, onBlocked, pNode->childNodes[0].get(), blocked);
            //forEachBlockedImpl(ac, onBlocked, pNode->childNodes[1].get(), blocked);
            pNode = pNode->childNodes[1].get();
        }
    }

    std::unique_ptr<Node> build(const Problem& problem,
                                const std::vector<uint64_t>& idxs) const
    {
        std::unique_ptr<Node> pNode(new Node);
        if (idxs.size() == 1)
        {
            uint64_t idx = idxs[0];
            Coord coord = problem.attendees[idx].coord;
            pNode->idx = idx;
            pNode->xs[0] = coord.x;
            pNode->xs[1] = coord.x;
            pNode->ys[0] = coord.y;
            pNode->ys[1] = coord.y;
        }
        else
        {
            double lowX = std::numeric_limits<double>::max();
            double lowY = std::numeric_limits<double>::max();
            double highX = std::numeric_limits<double>::lowest();
            double highY = std::numeric_limits<double>::lowest();
            for (uint64_t idx : idxs)
            {
                Coord coord = problem.attendees[idx].coord;
                lowX = std::min(lowX, coord.x);
                lowY = std::min(lowY, coord.y);
                highX = std::max(highX, coord.x);
                highY = std::max(highY, coord.y);
            }

            pNode->idx = (uint64_t)-1;
            pNode->xs[0] = lowX;
            pNode->ys[0] = lowY;
            pNode->xs[1] = highX;
            pNode->ys[1] = highY;

            std::vector<uint64_t> idxs0;
            std::vector<uint64_t> idxs1;

            if (lowX == highX && lowY == highY)
            {
                // All the same, split arbitrarily
                size_t half = idxs.size() / 2;
                idxs0.insert(idxs0.end(), idxs.begin(), idxs.begin() + half);
                idxs1.insert(idxs1.end(), idxs.begin() + half, idxs.end());
            }
            else if (highX - lowX >= highY - lowY)
            {
                // Split X
                double midX = lowX + (highX - lowX) * 0.5;
                if (midX == highX) midX = lowX;
                for (uint64_t idx : idxs)
                {
                    Coord coord = problem.attendees[idx].coord;
                    if (coord.x <= midX)
                    {
                        idxs0.push_back(idx);
                    }
                    else
                    {
                        idxs1.push_back(idx);
                    }
                }
            }
            else
            {
                // Split Y
                double midY = lowY + (highY - lowY) * 0.5;
                if (midY == highY) midY = lowY;
                for (uint64_t idx : idxs)
                {
                    Coord coord = problem.attendees[idx].coord;
                    if (coord.y <= midY)
                    {
                        idxs0.push_back(idx);
                    }
                    else
                    {
                        idxs1.push_back(idx);
                    }
                }
            }

            pNode->childNodes[0] = build(problem, idxs0);
            pNode->childNodes[1] = build(problem, idxs1);
        }
        return pNode;
    }

private:
    std::unique_ptr<Node> pRoot;
};

#endif
