#ifndef CHECKER_HPP
#define CHECKER_HPP

#include "Common.hpp"
#include <cmath>

class ObstructionChecker
{
public:
    ObstructionChecker(const Coord& attendeeCoord,
                       const Coord& musicianCoord_) :
        musicianCoord(musicianCoord_)
    {
        Coord ma = attendeeCoord - musicianCoord;
        double maL2 = calcL2(ma);
        maL = std::sqrt(maL2);
        parNorm = ma / maL;
        perpNorm = Coord(parNorm.y, -parNorm.x);
    }

    bool check(const Coord& obstructionCoord,
               double obstructionRadius) const
    {
        Coord mt = obstructionCoord - musicianCoord;
        double parL = calcDot(parNorm, mt);
        double perpL = calcDot(perpNorm, mt);
        return
            perpL > -obstructionRadius &&
            perpL < obstructionRadius &&
            parL > 0.0 &&
            parL < maL;
    }

private:
    Coord musicianCoord;
    double maL;
    Coord parNorm;
    Coord perpNorm;
};

enum class BlockedStatus : uint64_t
{
    Unblocked,
    Blocked,
    Unknown
};

class AttendeeChecker
{
public:
    AttendeeChecker(const Coord& musicianCoord_,
                    const Coord& obstructionCoord,
                    double obstructionRadius) :
        musicianCoord(musicianCoord_)
    {
        double obstructionR2 = obstructionRadius * obstructionRadius;
        double d2 = calcD2(musicianCoord, obstructionCoord);
        double val1 = obstructionR2 / d2;
        double val2 = obstructionRadius / d2 * std::sqrt(d2 - obstructionR2);
        Coord dv = musicianCoord - obstructionCoord;
        double val1X = val1 * dv.x;
        double val1Y = val1 * dv.y;
        double val2X = val2 * (-dv.y);
        double val2Y = val2 * dv.x;
        Coord pv1(val1X + val2X, val1Y + val2Y);
        Coord pv2(val1X - val2X, val1Y - val2Y);
        Coord p1 = obstructionCoord + pv1;
        Coord p2 = obstructionCoord + pv2;
        Coord pd1 = p1 - musicianCoord;
        Coord pd2 = p2 - musicianCoord;
        pd1Perp = Coord(pd1.y, -pd1.x);
        pd2Perp = Coord(-pd2.y, pd2.x);
        magicPoint = (p1 + p2) * 0.5;
        magicVec = musicianCoord - magicPoint;

        whichX1 = pd1Perp.x < 0.0 ? 1 : 0;
        whichY1 = pd1Perp.y < 0.0 ? 1 : 0;
        whichX2 = pd2Perp.x < 0.0 ? 1 : 0;
        whichY2 = pd2Perp.y < 0.0 ? 1 : 0;
        whichXM = magicVec.x < 0.0 ? 1 : 0;
        whichYM = magicVec.y < 0.0 ? 1 : 0;
    }

    bool check(const Coord& attendeeCoord) const
    {
        Coord ma = attendeeCoord - musicianCoord;
        Coord oa = attendeeCoord - magicPoint;
        bool blocked = false;
        if (calcDot(pd1Perp, ma) < 0.0 &&
            calcDot(pd2Perp, ma) < 0.0 &&
            calcDot(magicVec, oa) < 0.0)
        {
            blocked = true;
        }
        return blocked;
    }

    BlockedStatus checkRect(const double xs[2],
                            const double ys[2]) const
    {
        Coord corner1(xs[whichX1] - musicianCoord.x,
                      ys[whichY1] - musicianCoord.y);
        if (calcDot(pd1Perp, corner1) >= 0.0)
        {
            return BlockedStatus::Unblocked;
        }

        Coord corner2(xs[whichX2] - musicianCoord.x,
                      ys[whichY2] - musicianCoord.y);
        if (calcDot(pd2Perp, corner2) >= 0.0)
        {
            return BlockedStatus::Unblocked;
        }

        Coord cornerM(xs[whichXM] - magicPoint.x,
                      ys[whichYM] - magicPoint.y);
        if (calcDot(magicVec, cornerM) >= 0.0)
        {
            return BlockedStatus::Unblocked;
        }

#if 0
        Coord oppCorner1(xs[whichX1 ^ 1] - musicianCoord.x,
                         ys[whichY1 ^ 1] - musicianCoord.y);
        Coord oppCorner2(xs[whichX2 ^ 1] - musicianCoord.x,
                         ys[whichY2 ^ 1] - musicianCoord.y);
        Coord oppCornerM(xs[whichXM ^ 1] - magicPoint.x,
                         ys[whichYM ^ 1] - magicPoint.y);

        if (calcDot(pd1Perp, oppCorner1) < 0.0 &&
            calcDot(pd2Perp, oppCorner2) < 0.0 &&
            calcDot(magicVec, oppCornerM) < 0.0)
        {
            return BlockedStatus::Blocked;
        }
#endif

        return BlockedStatus::Unknown;
    }

    BlockedStatus checkRectFull(const double xs[2],
                                const double ys[2]) const
    {
        Coord corner1(xs[whichX1] - musicianCoord.x,
                      ys[whichY1] - musicianCoord.y);
        if (calcDot(pd1Perp, corner1) >= 0.0)
        {
            return BlockedStatus::Unblocked;
        }

        Coord corner2(xs[whichX2] - musicianCoord.x,
                      ys[whichY2] - musicianCoord.y);
        if (calcDot(pd2Perp, corner2) >= 0.0)
        {
            return BlockedStatus::Unblocked;
        }

        Coord cornerM(xs[whichXM] - magicPoint.x,
                      ys[whichYM] - magicPoint.y);
        if (calcDot(magicVec, cornerM) >= 0.0)
        {
            return BlockedStatus::Unblocked;
        }

#if 1
        Coord oppCorner1(xs[whichX1 ^ 1] - musicianCoord.x,
                         ys[whichY1 ^ 1] - musicianCoord.y);
        Coord oppCorner2(xs[whichX2 ^ 1] - musicianCoord.x,
                         ys[whichY2 ^ 1] - musicianCoord.y);
        Coord oppCornerM(xs[whichXM ^ 1] - magicPoint.x,
                         ys[whichYM ^ 1] - magicPoint.y);

        if (calcDot(pd1Perp, oppCorner1) < 0.0 &&
            calcDot(pd2Perp, oppCorner2) < 0.0 &&
            calcDot(magicVec, oppCornerM) < 0.0)
        {
            return BlockedStatus::Blocked;
        }
#endif

        return BlockedStatus::Unknown;
    }

private:
    Coord musicianCoord;
    Coord pd1Perp;
    Coord pd2Perp;
    Coord magicPoint;
    Coord magicVec;
    size_t whichX1;
    size_t whichY1;
    size_t whichX2;
    size_t whichY2;
    size_t whichXM;
    size_t whichYM;
};

#endif
