#include "DrawProblem.hpp"
#include "Graphics.hpp"
#include "Problem.hpp"

using std::string;

namespace
{
    uint64_t calcHash(uint64_t key)
    {
#if 1
        uint64_t hash = key;
        hash ^= hash >> 33;
        hash *= 0xff51afd7ed558ccd;
        hash ^= hash >> 33;
        hash *= 0xc4ceb9fe1a85ec53;
        hash ^= hash >> 33;
#endif

#if 0
        uint64_t hash = key;
        hash ^= hash >> 30;
        hash *= 0xbf58476d1ce4e5b9;
        hash ^= hash >> 27;
        hash *= 0x94d049bb133111eb;
        hash ^= hash >> 31;
#endif

        return hash;
    }

    uint32_t getInstrumentColor(uint64_t instrument)
    {
        uint64_t hash64 = calcHash(instrument);
        uint32_t hash32 = hash64 ^ (hash64 >> 32);
        return ((uint32_t)hash32 & 0xffffff) | 0x808080;
    }

    uint32_t getSilentColor(uint32_t color)
    {
        uint32_t r = (color >> 16) & 0xff;
        uint32_t g = (color >> 8) & 0xff;
        uint32_t b = color & 0xff;
        r /= 2;
        g /= 2;
        b /= 2;
        return (r << 16) | (g << 8) | b;
    }
}

void drawProblem(Graphics& gr,
                 const Problem& problem,
                 const Solution& solution,
                 const string& text)
{
    uint32_t width = gr.getWidth();
    uint32_t height = gr.getHeight();

    gr.setFont(FontSpacing::Monospace, FontFace::SansSerif,
               18, false, false);

    gr.fillRect(0, 0, width, height, 0x404040);

    if (!text.empty())
    {
        gr.drawText(2, 0,
                    text.c_str(),
                    0xffffff,
                    HAlign::Left, VAlign::Top);
    }

    uint32_t margin = 40;
    if (width <= margin * 2 || height <= margin * 2) return;

    double scale;
    if (problem.roomWidth * (double)(height - margin * 2) >= (double)(width - margin * 2) * problem.roomHeight)
    {
        scale = (double)(width - margin * 2) / problem.roomWidth;
    }
    else
    {
        scale = (double)(height - margin * 2) / problem.roomHeight;
    }

    int32_t cornerX = (int32_t)(width / 2) - (int32_t)(problem.roomWidth * 0.5 * scale);
    int32_t cornerY = (int32_t)(height / 2) + (int32_t)(problem.roomHeight * 0.5 * scale);

    auto convertDist = [&](double worldWidth) -> int32_t
    {
        return (int32_t)worldWidth * scale;
    };
    auto convertX = [&](double worldX) -> int32_t
    {
        return cornerX + (int32_t)worldX * scale;
    };
    auto convertY = [&](double worldY) -> int32_t
    {
        return cornerY - (int32_t)worldY * scale;
    };

    gr.fillRect(convertX(0.0),
                convertY(problem.roomHeight),
                convertDist(problem.roomWidth),
                convertDist(problem.roomHeight),
                0x000000);

    if (problem.stageWidth > 10 && problem.stageHeight > 10)
    {
        gr.fillRect(convertX(problem.stageX + 5),
                    convertY(problem.stageY + problem.stageHeight - 5),
                    convertDist(problem.stageWidth - 10),
                    convertDist(problem.stageHeight - 10),
                    0x0000ff);
    }

    for (auto& pillar : problem.pillars)
    {
        Coord coord = pillar.coord;
        double worldDiameter = pillar.radius * 2.0;
        int32_t diameter = convertDist(worldDiameter);
        if (diameter < 2) diameter = 2;
        gr.fillCircle(convertX(coord.x) - diameter / 2,
                      convertY(coord.y) - diameter / 2,
                      diameter,
                      diameter,
                      0x804000);
    }

    double worldDiameter = 10.0;
    int32_t diameter = convertDist(worldDiameter);
    if (diameter < 2) diameter = 2;

    for (auto& attendee : problem.attendees)
    {
        Coord coord = attendee.coord;
        gr.fillCircle(convertX(coord.x) - diameter / 2,
                      convertY(coord.y) - diameter / 2,
                      diameter,
                      diameter,
                      0xffffff);
    }

    for (uint64_t musicianIdx = 0; musicianIdx < solution.placements.size(); musicianIdx++)
    {
        uint64_t instrument = problem.musicians[musicianIdx];
        uint32_t color = getInstrumentColor(instrument);
        if (!solution.volumes.empty() && solution.volumes[musicianIdx] == 0.0)
        {
            color = getSilentColor(color);
        }
        Coord coord = solution.placements[musicianIdx];
        gr.fillCircle(convertX(coord.x) - diameter / 2,
                      convertY(coord.y) - diameter / 2,
                      diameter,
                      diameter,
                      color);
    }
}
