#ifndef FASTSCORE_HPP
#define FASTSCORE_HPP

#include "Common.hpp"
#include "Problem.hpp"
#include "AttendeeIndex.hpp"
#include <cmath>

class FastScore
{
public:
    FastScore(const AttendeeIndex& index_) : index(index_) {}

    void init(const Problem& problem, const Solution& solution)
    {
        uint64_t attendeeCount = problem.attendees.size();
        uint64_t musicianCount = problem.musicians.size();
        blockedCounts.resize(attendeeCount * musicianCount);
        closenesses.resize(musicianCount);
        impacts.resize(musicianCount);

        bool useCloseness = !problem.pillars.empty();

        for (uint64_t musicianIdx = 0; musicianIdx < musicianCount; musicianIdx++)
        {
            Coord musicianCoord = solution.placements[musicianIdx];

            double closeness = 1.0;
            if (useCloseness)
            {
                for (uint64_t otherMusicianIdx : problem.instruments[problem.musicians[musicianIdx]])
                {
                    if (otherMusicianIdx == musicianIdx) continue;
                    Coord otherMusicianCoord = solution.placements[otherMusicianIdx];
                    double d2 = calcD2(musicianCoord, otherMusicianCoord);
                    double d = std::sqrt(d2);
                    closeness += 1.0 / d;
                }
                closenesses[musicianIdx] = closeness;
            }

            for (uint64_t attendeeIdx = 0; attendeeIdx < attendeeCount; attendeeIdx++)
            {
                blockedCounts[musicianIdx * attendeeCount + attendeeIdx] = 0;
            }

            auto onBlocked = [&](uint64_t attendeeIdx)
            {
                blockedCounts[musicianIdx * attendeeCount + attendeeIdx]++;
            };

            for (uint64_t testIdx = 0; testIdx < musicianCount; testIdx++)
            {
                if (testIdx == musicianIdx) continue;
                Coord testCoord = solution.placements[testIdx];
                index.forEachBlocked(musicianCoord, testCoord, 5.0, onBlocked);
            }

            for (auto& pillar : problem.pillars)
            {
                index.forEachBlocked(musicianCoord, pillar.coord, pillar.radius, onBlocked);
            }

            double musicianImpact = 0.0;

            for (uint64_t attendeeIdx = 0; attendeeIdx < attendeeCount; attendeeIdx++)
            {
                uint32_t blockedCount = blockedCounts[musicianIdx * attendeeCount + attendeeIdx];

                if (blockedCount != 0) continue;

                auto& attendee = problem.attendees[attendeeIdx];
                Coord attendeeCoord = attendee.coord;

                double d2 = calcD2(attendeeCoord, musicianCoord);
                double taste = attendee.tastes[problem.musicians[musicianIdx]];

                double impact = 10.0 * std::ceil(1000000.0 * taste / d2);
                if (useCloseness)
                {
                    impact = std::ceil(closeness * impact);
                }
                musicianImpact += impact;
            }

            impacts[musicianIdx] = musicianImpact;
        }

        blockedCountsBackup = blockedCounts;
        closenessesBackup = closenesses;
        impactsBackup = impacts;
    }

    void update(const Problem& problem,
                Solution& solution,
                uint64_t musicianIdx,
                Coord newMusicianCoord)
    {
        uint64_t attendeeCount = problem.attendees.size();
        uint64_t musicianCount = problem.musicians.size();

        Coord oldMusicianCoord = solution.placements[musicianIdx];
        solution.placements[musicianIdx] = newMusicianCoord;

        uint64_t instrument = problem.musicians[musicianIdx];
        bool useCloseness = !problem.pillars.empty();

        if (useCloseness)
        {
            for (uint64_t musicianIdxA : problem.instruments[instrument])
            {
                if (musicianIdxA == musicianIdx) continue;
                Coord musicianCoordA = solution.placements[musicianIdxA];

                double closeness = 1.0;
                for (uint64_t musicianIdxB : problem.instruments[instrument])
                {
                    if (musicianIdxB == musicianIdxA) continue;
                    Coord musicianCoordB = solution.placements[musicianIdxB];
                    double d2 = calcD2(musicianCoordA, musicianCoordB);
                    double d = std::sqrt(d2);
                    closeness += 1.0 / d;
                }

                closenesses[musicianIdxA] = closeness;

                double musicianImpact = 0.0;

                for (uint64_t attendeeIdx = 0; attendeeIdx < attendeeCount; attendeeIdx++)
                {
                    uint32_t blockedCount = blockedCounts[musicianIdxA * attendeeCount + attendeeIdx];

                    if (blockedCount != 0) continue;

                    auto& attendee = problem.attendees[attendeeIdx];
                    Coord attendeeCoord = attendee.coord;

                    double d2 = calcD2(attendeeCoord, musicianCoordA);
                    double taste = attendee.tastes[instrument];

                    double impact = 10.0 * std::ceil(1000000.0 * taste / d2);
                    musicianImpact += std::ceil(closeness * impact);
                }

                impacts[musicianIdxA] = musicianImpact;
            }
        }

        double closeness = 1.0;
        if (useCloseness)
        {
            for (uint64_t musicianIdxB : problem.instruments[instrument])
            {
                if (musicianIdxB == musicianIdx) continue;
                Coord musicianCoordB = solution.placements[musicianIdxB];
                double d2 = calcD2(newMusicianCoord, musicianCoordB);
                double d = std::sqrt(d2);
                closeness += 1.0 / d;
            }
            closenesses[musicianIdx] = closeness;
        }

        for (uint64_t attendeeIdx = 0; attendeeIdx < attendeeCount; attendeeIdx++)
        {
            blockedCounts[musicianIdx * attendeeCount + attendeeIdx] = 0;
        }

        auto onBlocked = [&](uint64_t attendeeIdx)
        {
            blockedCounts[musicianIdx * attendeeCount + attendeeIdx]++;
        };

        for (uint64_t testIdx = 0; testIdx < musicianCount; testIdx++)
        {
            if (testIdx == musicianIdx) continue;
            Coord testCoord = solution.placements[testIdx];
            index.forEachBlocked(newMusicianCoord, testCoord, 5.0, onBlocked);
        }

        for (auto& pillar : problem.pillars)
        {
            index.forEachBlocked(newMusicianCoord, pillar.coord, pillar.radius, onBlocked);
        }

        double musicianImpact = 0.0;

        for (uint64_t attendeeIdx = 0; attendeeIdx < attendeeCount; attendeeIdx++)
        {
            uint32_t blockedCount = blockedCounts[musicianIdx * attendeeCount + attendeeIdx];

            if (blockedCount != 0) continue;

            auto& attendee = problem.attendees[attendeeIdx];
            Coord attendeeCoord = attendee.coord;

            double d2 = calcD2(attendeeCoord, newMusicianCoord);
            double taste = attendee.tastes[instrument];

            double impact = 10.0 * std::ceil(1000000.0 * taste / d2);
            if (useCloseness)
            {
                impact = std::ceil(closeness * impact);
            }
            musicianImpact += impact;
        }

        impacts[musicianIdx] = musicianImpact;

        for (uint64_t testIdx = 0; testIdx < musicianCount; testIdx++)
        {
            if (testIdx == musicianIdx) continue;
            Coord testCoord = solution.placements[testIdx];

            double musicianImpact = impacts[testIdx];

            {
                auto onBlocked = [&](uint64_t attendeeIdx)
                {
                    uint32_t blockedCount = blockedCounts[testIdx * attendeeCount + attendeeIdx];
                    blockedCount--;
                    blockedCounts[testIdx * attendeeCount + attendeeIdx] = blockedCount;
                    if (blockedCount == 0)
                    {
                        auto& attendee = problem.attendees[attendeeIdx];
                        Coord attendeeCoord = attendee.coord;

                        double d2 = calcD2(attendeeCoord, testCoord);
                        double taste = attendee.tastes[problem.musicians[testIdx]];

                        double impact = 10.0 * std::ceil(1000000.0 * taste / d2);
                        if (useCloseness)
                        {
                            impact = std::ceil(closenesses[testIdx] * impact);
                        }
                        musicianImpact += impact;
                    }
                };

                index.forEachBlocked(testCoord, oldMusicianCoord, 5.0, onBlocked);
            }

            {
                auto onBlocked = [&](uint64_t attendeeIdx)
                {
                    uint32_t blockedCount = blockedCounts[testIdx * attendeeCount + attendeeIdx];
                    blockedCount++;
                    blockedCounts[testIdx * attendeeCount + attendeeIdx] = blockedCount;
                    if (blockedCount == 1)
                    {
                        auto& attendee = problem.attendees[attendeeIdx];
                        Coord attendeeCoord = attendee.coord;

                        double d2 = calcD2(attendeeCoord, testCoord);
                        double taste = attendee.tastes[problem.musicians[testIdx]];

                        double impact = 10.0 * std::ceil(1000000.0 * taste / d2);
                        if (useCloseness)
                        {
                            impact = std::ceil(closenesses[testIdx] * impact);
                        }
                        musicianImpact -= impact;
                    }
                };

                index.forEachBlocked(testCoord, newMusicianCoord, 5.0, onBlocked);
            }

            impacts[testIdx] = musicianImpact;
        }
    }

    int64_t setVolumes(const Problem& problem, Solution& solution)
    {
        uint64_t musicianCount = problem.musicians.size();
        solution.volumes.resize(musicianCount);

        double score = 0.0;

        for (uint64_t musicianIdx = 0; musicianIdx < musicianCount; musicianIdx++)
        {
            if (impacts[musicianIdx] > 0.0)
            {
                solution.volumes[musicianIdx] = 10.0;
                score += impacts[musicianIdx];
            }
            else
            {
                solution.volumes[musicianIdx] = 0.0;
            }
        }

        return (int64_t)score;
    }

    void commit()
    {
        blockedCountsBackup = blockedCounts;
        closenessesBackup = closenesses;
        impactsBackup = impacts;
    }

    void rollback()
    {
        blockedCounts = blockedCountsBackup;
        closenesses = closenessesBackup;
        impacts = impactsBackup;
    }

private:
    const AttendeeIndex& index;
    std::vector<uint32_t> blockedCounts;
    std::vector<double> closenesses;
    std::vector<double> impacts;
    std::vector<uint32_t> blockedCountsBackup;
    std::vector<double> closenessesBackup;
    std::vector<double> impactsBackup;
};

#endif
