#include "Problem.hpp"
#include "Json.hpp"
#include "JsonUtils.hpp"
#include "FileUtils.hpp"
#include "StringUtils.hpp"

using std::string;
using std::vector;

namespace J = rapidjson;

namespace
{
    void parseMusician(const J::Value& jMusician,
                       uint64_t* pMusician)
    {
        auto& musician = *pMusician;

        musician = JGetUint64(jMusician);

        if (musician > 1000000)
        {
            throw JException("jMusician: instrument number too high");
        }
    }

    void parseMusicians(const J::Value& jMusicians,
                        vector<uint64_t>* pMusicians)
    {
        auto& musicians = *pMusicians;

        JCheckArray(jMusicians);

        musicians.clear();
        musicians.reserve(jMusicians.Size());
        for (auto& jMusician : jMusicians.GetArray())
        {
            auto& musician = musicians.emplace_back();
            parseMusician(jMusician, &musician);
        }
    }

    void parseTaste(const J::Value& jTaste,
                    double* pTaste)
    {
        auto& taste = *pTaste;

        taste = JGetDouble(jTaste);
    }

    void parseTastes(const J::Value& jTastes,
                     vector<double>* pTastes)
    {
        auto& tastes = *pTastes;

        JCheckArray(jTastes);

        tastes.clear();
        tastes.reserve(jTastes.Size());
        for (auto& jTaste : jTastes.GetArray())
        {
            auto& taste = tastes.emplace_back();
            parseTaste(jTaste, &taste);
        }
    }

    void parseAttendee(const J::Value& jAttendee,
                       Attendee* pAttendee)
    {
        auto& attendee = *pAttendee;

        JCheckObject(jAttendee);

        attendee.coord.x = JObjectGetDouble(jAttendee, "x");
        attendee.coord.y = JObjectGetDouble(jAttendee, "y");

        {
            auto& jTastes = JObjectGetArray(jAttendee, "tastes");
            parseTastes(jTastes, &attendee.tastes);
        }
    }

    void parseAttendees(const J::Value& jAttendees,
                        vector<Attendee>* pAttendees)
    {
        auto& attendees = *pAttendees;

        JCheckArray(jAttendees);

        attendees.clear();
        attendees.reserve(jAttendees.Size());
        for (auto& jAttendee : jAttendees.GetArray())
        {
            auto& attendee = attendees.emplace_back();
            parseAttendee(jAttendee, &attendee);
        }
    }

    void parsePillar(const J::Value& jPillar,
                     Pillar* pPillar)
    {
        auto& pillar = *pPillar;

        JCheckObject(jPillar);

        {
            auto& jCenter = JObjectGetArray(jPillar, "center");
            if (jCenter.Size() != 2)
            {
                throw JException("jCenter: incorrect size");
            }

            {
                auto& jCenterX = jCenter[0];
                pillar.coord.x = JGetDouble(jCenterX);
            }

            {
                auto& jCenterY = jCenter[1];
                pillar.coord.y = JGetDouble(jCenterY);
            }
        }


        pillar.radius = JObjectGetDouble(jPillar, "radius");
    }

    void parsePillars(const J::Value& jPillars,
                      vector<Pillar>* pPillars)
    {
        auto& pillars = *pPillars;

        JCheckArray(jPillars);

        pillars.clear();
        pillars.reserve(jPillars.Size());
        for (auto& jPillar : jPillars.GetArray())
        {
            auto& pillar = pillars.emplace_back();
            parsePillar(jPillar, &pillar);
        }
    }

    void parseProblem(const J::Value& jProblem,
                      Problem* pProblem)
    {
        auto& problem = *pProblem;

        JCheckObject(jProblem);

        problem.roomWidth = JObjectGetDouble(jProblem, "room_width");
        problem.roomHeight = JObjectGetDouble(jProblem, "room_height");
        problem.stageWidth = JObjectGetDouble(jProblem, "stage_width");
        problem.stageHeight = JObjectGetDouble(jProblem, "stage_height");

        {
            auto& jStageBottomLeft = JObjectGetArray(jProblem, "stage_bottom_left");
            if (jStageBottomLeft.Size() != 2)
            {
                throw JException("jStageBottomLeft: incorrect size");
            }

            {
                auto& jStageX = jStageBottomLeft[0];
                problem.stageX = JGetDouble(jStageX);
            }

            {
                auto& jStageY = jStageBottomLeft[1];
                problem.stageY = JGetDouble(jStageY);
            }
        }

        {
            auto& jMusicians = JObjectGetArray(jProblem, "musicians");
            parseMusicians(jMusicians, &problem.musicians);
        }

        {
            auto& jAttendees = JObjectGetArray(jProblem, "attendees");
            parseAttendees(jAttendees, &problem.attendees);
        }

        if (jProblem.HasMember("pillars"))
        {
            auto& jPillars = JObjectGetArray(jProblem, "pillars");
            parsePillars(jPillars, &problem.pillars);
        }

        for (uint64_t musicianIdx = 0; musicianIdx < problem.musicians.size(); musicianIdx++)
        {
            uint64_t instrument = problem.musicians[musicianIdx];
            if (instrument >= problem.instruments.size())
            {
                problem.instruments.resize(instrument + 1);
            }
            problem.instruments[instrument].push_back(musicianIdx);
        }
        for (auto& musicians : problem.instruments)
        {
            if (musicians.empty())
            {
                throw JException("Unused instrument");
            }
        }
    }

    void parseCoord(const J::Value& jCoord,
                    Coord* pCoord)
    {
        auto& coord = *pCoord;

        JCheckObject(jCoord);

        coord.x = JObjectGetDouble(jCoord, "x");
        coord.y = JObjectGetDouble(jCoord, "y");
    }

    void parseCoords(const J::Value& jCoords,
                     vector<Coord>* pCoords)
    {
        auto& coords = *pCoords;

        JCheckArray(jCoords);

        coords.clear();
        coords.reserve(jCoords.Size());
        for (auto& jCoord : jCoords.GetArray())
        {
            auto& coord = coords.emplace_back();
            parseCoord(jCoord, &coord);
        }
    }

    void parseVolume(const J::Value& jVolume,
                     double* pVolume)
    {
        auto& volume = *pVolume;

        volume = JGetDouble(jVolume);
    }

    void parseVolumes(const J::Value& jVolumes,
                      vector<double>* pVolumes)
    {
        auto& volumes = *pVolumes;

        JCheckArray(jVolumes);

        volumes.clear();
        volumes.reserve(jVolumes.Size());
        for (auto& jVolume : jVolumes.GetArray())
        {
            auto& volume = volumes.emplace_back();
            parseVolume(jVolume, &volume);
        }
    }

    void parseSolution(const J::Value& jSolution,
                       Solution* pSolution)
    {
        auto& solution = *pSolution;

        JCheckObject(jSolution);

        {
            auto& jPlacements = JObjectGetArray(jSolution, "placements");
            parseCoords(jPlacements, &solution.placements);
        }

        if (jSolution.HasMember("volumes"))
        {
            auto& jVolumes = JObjectGetArray(jSolution, "volumes");
            parseVolumes(jVolumes, &solution.volumes);
        }
    }

    template<typename Writer>
    void formatProblem(const Problem& problem,
                       Writer& w)
    {
        w.StartObject();

        w.Key("room_width");
        w.Double(problem.roomWidth);

        w.Key("room_height");
        w.Double(problem.roomHeight);

        w.Key("stage_width");
        w.Double(problem.stageWidth);

        w.Key("stage_height");
        w.Double(problem.stageHeight);

        w.Key("stage_bottom_left");
        w.StartArray();
        w.Double(problem.stageX);
        w.Double(problem.stageY);
        w.EndArray();

        w.Key("musicians");
        w.StartArray();
        for (auto& musician : problem.musicians)
        {
            w.Uint64(musician);
        }
        w.EndArray();

        w.Key("attendees");
        w.StartArray();
        for (auto& attendee : problem.attendees)
        {
            w.StartObject();

            w.Key("x");
            w.Double(attendee.coord.x);

            w.Key("y");
            w.Double(attendee.coord.y);

            w.Key("tastes");
            w.StartArray();
            for (auto& taste : attendee.tastes)
            {
                w.Double(taste);
            }
            w.EndArray();

            w.EndObject();
        }
        w.EndArray();

        w.Key("pillars");
        w.StartArray();
        for (auto& pillar : problem.pillars)
        {
            w.StartObject();

            w.Key("center");
            w.StartArray();
            w.Double(pillar.coord.x);
            w.Double(pillar.coord.y);
            w.EndArray();

            w.Key("radius");
            w.Double(pillar.radius);

            w.EndObject();
        }
        w.EndArray();

        w.EndObject();
    }

    template<typename Writer>
    void formatSolution(const Solution& solution,
                        Writer& w)
    {
        w.StartObject();

        w.Key("placements");
        w.StartArray();
        for (auto& coord : solution.placements)
        {
            w.StartObject();

            w.Key("x");
            w.Double(coord.x);

            w.Key("y");
            w.Double(coord.y);

            w.EndObject();
        }
        w.EndArray();

        if (!solution.volumes.empty())
        {
            w.Key("volumes");
            w.StartArray();
            for (auto& volume : solution.volumes)
            {
                w.Double(volume);
            }
            w.EndArray();
        }

        w.EndObject();
    }
}

bool readProblem(uint64_t id,
                 string* pProblemData)
{
    string fileName = strprintf("../problems/%" PRIu64 "", id);
    return readFile(fileName, pProblemData);
}

bool writeProblem(uint64_t id,
                  const string& problemData)
{
    string fileName = strprintf("../problems/%" PRIu64 "", id);
    return writeFile(fileName, problemData);
}

bool parseProblem(const string& problemData,
                  Problem* pProblem,
                  string* pMsg)
{
    try
    {
        J::Document jProblem;
        jProblem.Parse(problemData);
        JCheckParse(jProblem);

        parseProblem(jProblem, pProblem);
    }
    catch (const JException& e)
    {
        if (pMsg) *pMsg = e.m_msg;
        return false;
    }

    return true;
}

bool parseProblem(uint64_t id,
                  Problem* pProblem,
                  string* pMsg)
{
    string problemData;
    if (!readProblem(id, &problemData))
    {
        if (pMsg) *pMsg = strprintf("Error reading problem %" PRIu64 "", id);
        return false;
    }

    return parseProblem(problemData, pProblem, pMsg);
}

void formatProblem(const Problem& problem,
                   string* pProblemData)
{
    J::StringBuffer buffer;
    J::Writer<J::StringBuffer> w(buffer);

    formatProblem(problem, w);

    *pProblemData = buffer.GetString();
}

bool readSolution(uint64_t id,
                  string* pSolutionData)
{
    string fileName = strprintf("../solutions/%" PRIu64 "", id);
    return readFile(fileName, pSolutionData);
}

bool writeSolution(uint64_t id,
                   const string& solutionData)
{
    string fileName = strprintf("../solutions/%" PRIu64 "", id);
    return writeFile(fileName, solutionData);
}

bool parseSolution(const string& solutionData,
                   Solution* pSolution,
                   string* pMsg)
{
    try
    {
        J::Document jSolution;
        jSolution.Parse(solutionData);
        JCheckParse(jSolution);

        parseSolution(jSolution, pSolution);
    }
    catch (const JException& e)
    {
        if (pMsg) *pMsg = e.m_msg;
        return false;
    }

    return true;
}

bool parseSolution(uint64_t id,
                   Solution* pSolution,
                   string* pMsg)
{
    string solutionData;
    if (!readSolution(id, &solutionData))
    {
        if (pMsg) *pMsg = strprintf("Error reading solution %" PRIu64 "", id);
        return false;
    }

    return parseSolution(solutionData, pSolution, pMsg);
}

void formatSolution(const Solution& solution,
                    string* pSolutionData)
{
    J::StringBuffer buffer;
    J::Writer<J::StringBuffer> w(buffer);

    formatSolution(solution, w);

    *pSolutionData = buffer.GetString();
}
