#include "Protocol.hpp"
#include "Http.hpp"
#include "FileUtils.hpp"
#include "StringUtils.hpp"
#include "Json.hpp"
#include "JsonUtils.hpp"

using std::string;
using std::vector;

namespace J = rapidjson;

namespace Protocol
{
    constexpr const char* buildId = "6yeVot_EFqxS38t_pKJ1E";

    void init()
    {
        Http::init();
    }

    void cleanup()
    {
        Http::cleanup();
    }

    bool getAPIKey(string* pKey)
    {
        string tmpKey;
        if (!readFile("../../../apikey.txt", &tmpKey))
        {
            return false;
        }

        string key;
        for (char ch : tmpKey)
        {
            if (ch >= 33 && ch <= 126)
            {
                key.push_back(ch);
            }
        }
        if (key.empty())
        {
            return false;
        }

        if (pKey) *pKey = std::move(key);
        return true;
    }

    bool getAPIKeyHeader(vector<string>* pHeaders,
                         std::string* pMsg = nullptr)
    {
        string key;
        if (!getAPIKey(&key))
        {
            if (pMsg) *pMsg = "Error reading API key";
            return false;
        }

        pHeaders->push_back("Authorization: Bearer " + key);
        return true;
    }

    void getUserAgentHeader(vector<string>* pHeaders)
    {
        pHeaders->push_back("User-Agent: libcurl (Frictionless Bananas ICFPC 2023 tools)");
    }

    bool getProblem(uint64_t id,
                    string* pProblemData,
                    string* pMsg)
    {
        vector<string> headers;
        getUserAgentHeader(&headers);

        string url = strprintf("https://cdn.icfpcontest.com/problems/%" PRIu64 ".json", id);

        string response;
        if (!Http::get(url, &headers, &response, pMsg))
        {
            return false;
        }

        if (pProblemData) *pProblemData = std::move(response);
        return true;
    }

    bool getScoreboard(string* pScoreboardData,
                       string* pMsg)
    {
        vector<string> headers;
        if (!getAPIKeyHeader(&headers, pMsg)) return false;
        getUserAgentHeader(&headers);

        string url = "https://api.icfpcontest.com/scoreboard";

        string response;
        if (!Http::get(url, &headers, &response, pMsg))
        {
            return false;
        }

        if (pScoreboardData) *pScoreboardData = std::move(response);
        return true;
    }

    bool getUserboard(string* pUserboardData,
                      string* pMsg)
    {
        vector<string> headers;
        if (!getAPIKeyHeader(&headers, pMsg)) return false;
        getUserAgentHeader(&headers);

        string url = "https://api.icfpcontest.com/userboard";

        string response;
        if (!Http::get(url, &headers, &response, pMsg))
        {
            return false;
        }

        if (pUserboardData) *pUserboardData = std::move(response);
        return true;
    }

    bool submit(uint64_t id,
                const string& solutionData,
                string* pSubmissionId,
                string* pMsg)
    {
        vector<string> headers;
        if (!getAPIKeyHeader(&headers, pMsg)) return false;
        getUserAgentHeader(&headers);

        string url = "https://api.icfpcontest.com/submission";

        string request;

        {
            J::StringBuffer buffer;
            J::Writer<J::StringBuffer> w(buffer);

            w.StartObject();

            w.Key("problem_id");
            w.Uint64(id);

            w.Key("contents");
            w.String(solutionData);

            w.EndObject();

            request = buffer.GetString();
        }

        string response;
        if (!Http::post(url, request, "application/json", &headers, &response, pMsg))
        {
            return false;
        }

        string submissionId;
        try
        {
            J::Document jSubmissionId;
            jSubmissionId.Parse(response);
            JCheckParse(jSubmissionId);
            submissionId = JGetString(jSubmissionId);
        }
        catch (const JException& e)
        {
            if (pMsg) *pMsg = e.m_msg;
            return false;
        }

        if (pSubmissionId) *pSubmissionId = std::move(submissionId);
        return true;
    }

    bool updateUsername(const string& username,
                        string* pMsg)
    {
        vector<string> headers;
        if (!getAPIKeyHeader(&headers, pMsg)) return false;
        getUserAgentHeader(&headers);

        string url = "https://api.icfpcontest.com/user/update_username";

        string request;

        {
            J::StringBuffer buffer;
            J::Writer<J::StringBuffer> w(buffer);

            w.StartObject();

            w.Key("username");
            w.String(username);

            w.EndObject();

            request = buffer.GetString();
        }

        string response;
        if (!Http::post(url, request, "application/json", &headers, &response, pMsg))
        {
            return false;
        }

        return true;
    }
}
