#include "Scores.hpp"
#include "FileUtils.hpp"
#include "Json.hpp"
#include "JsonUtils.hpp"
#include "ProblemCount.hpp"

using std::string;
using std::vector;

namespace J = rapidjson;

namespace
{
    void parseScoreboardEntry(const J::Value& jEntry,
                              ScoreboardEntry* pEntry)
    {
        auto& entry = *pEntry;

        JCheckObject(jEntry);

        entry.username = JObjectGetString(jEntry, "username");
        entry.score = JObjectGetDouble(jEntry, "score");
    }

    void parseScoreboardEntries(const J::Value& jEntries,
                                vector<ScoreboardEntry>* pEntries)
    {
        auto& entries = *pEntries;

        JCheckArray(jEntries);

        entries.clear();
        entries.reserve(jEntries.Size());
        for (auto& jEntry : jEntries.GetArray())
        {
            auto& entry = entries.emplace_back();
            parseScoreboardEntry(jEntry, &entry);
        }
    }

    void parseScoreboardFile(const J::Value& jScoreboardFile,
                             Scoreboard* pScoreboard)
    {
        auto& scoreboard = *pScoreboard;

        JCheckObject(jScoreboardFile);

        scoreboard.frozen = JObjectGetBool(jScoreboardFile, "frozen");

        {
            auto& jEntries = JObjectGetArray(jScoreboardFile, "scoreboard");
            parseScoreboardEntries(jEntries, &scoreboard.entries);
        }

        scoreboard.updatedAt = JObjectGetString(jScoreboardFile, "updated_at");
    }
}

bool readScoreboard(string* pScoreboardData)
{
    string fileName = "../data/scoreboard";
    return readFile(fileName, pScoreboardData);
}

bool writeScoreboard(const string& scoreboardData)
{
    string fileName = "../data/scoreboard";
    return writeFile(fileName, scoreboardData);
}

bool parseScoreboard(const string& scoreboardData,
                     Scoreboard* pScoreboard,
                     string* pMsg)
{
    try
    {
        J::Document jScoreboardFile;
        jScoreboardFile.Parse(scoreboardData);
        JCheckParse(jScoreboardFile);

        parseScoreboardFile(jScoreboardFile, pScoreboard);
    }
    catch (const JException& e)
    {
        if (pMsg) *pMsg = e.m_msg;
        return false;
    }

    return true;
}

bool parseScoreboard(Scoreboard* pScoreboard,
                     string* pMsg)
{
    string scoreboardData;
    if (!readScoreboard(&scoreboardData))
    {
        if (pMsg) *pMsg = "Error reading scoreboard";
        return false;
    }

    return parseScoreboard(scoreboardData, pScoreboard, pMsg);
}

namespace
{
    void parseUserboardEntry(const J::Value& jEntry,
                             UserboardEntry* pEntry)
    {
        auto& entry = *pEntry;

        if (jEntry.IsNull())
        {
            entry.valid = false;
            entry.score = 0.0;
        }
        else
        {
            entry.valid = true;
            entry.score = JGetDouble(jEntry);
        }
    }

    void parseUserboardEntries(const J::Value& jEntries,
                               vector<UserboardEntry>* pEntries)
    {
        auto& entries = *pEntries;

        JCheckArray(jEntries);

        entries.clear();
        entries.reserve(jEntries.Size());
        for (auto& jEntry : jEntries.GetArray())
        {
            auto& entry = entries.emplace_back();
            parseUserboardEntry(jEntry, &entry);
        }
    }

    void parseUserboard(const J::Value& jUserboard,
                        Userboard* pUserboard)
    {
        auto& userboard = *pUserboard;

        JCheckObject(jUserboard);

        {
            auto& jEntries = JObjectGetArray(jUserboard, "problems");
            parseUserboardEntries(jEntries, &userboard.entries);
        }
    }

    void parseUserboardFile(const J::Value& jUserboardFile,
                            Userboard* pUserboard)
    {
        JCheckObject(jUserboardFile);

        {
            auto& jUserboard = JObjectGetObject(jUserboardFile, "Success");
            parseUserboard(jUserboard, pUserboard);
        }
    }
}

bool readUserboard(string* pUserboardData)
{
    string fileName = "../data/userboard";
    return readFile(fileName, pUserboardData);
}

bool writeUserboard(const string& userboardData)
{
    string fileName = "../data/userboard";
    return writeFile(fileName, userboardData);
}

bool parseUserboard(const string& userboardData,
                    Userboard* pUserboard,
                    string* pMsg)
{
    try
    {
        J::Document jUserboardFile;
        jUserboardFile.Parse(userboardData);
        JCheckParse(jUserboardFile);

        parseUserboardFile(jUserboardFile, pUserboard);
    }
    catch (const JException& e)
    {
        if (pMsg) *pMsg = e.m_msg;
        return false;
    }

    return true;
}

bool parseUserboard(Userboard* pUserboard,
                    string* pMsg)
{
    string userboardData;
    if (!readUserboard(&userboardData))
    {
        if (pMsg) *pMsg = "Error reading userboard";
        return false;
    }

    return parseUserboard(userboardData, pUserboard, pMsg);
}

void getMyScores(const Userboard& userboard,
                 vector<int64_t>* pMyScores)
{
    vector<int64_t> myScores(problemCount + 1, 0.0);
    for (uint64_t id = 1; id <= problemCount; id++)
    {
        if (id - 1 < userboard.entries.size())
        {
            auto& entry = userboard.entries[id - 1];
            if (entry.valid)
            {
                myScores[id] = (int64_t)entry.score;
            }
        }
    }

    if (pMyScores) *pMyScores = std::move(myScores);
}
