#include "Common.hpp"
#include "ProblemCount.hpp"
#include "ParseUtils.hpp"
#include "Problem.hpp"
#include "IsValid.hpp"
#include "CalcScore.hpp"
#include "Best.hpp"
#include "Xoshiro.hpp"
#include "TimeUtils.hpp"
#include "ProcessId.hpp"
#include "FastRandom.hpp"
#include "FormatUtils.hpp"
#include "FastScore.hpp"
#include "DebugGraphics.hpp"
#include "StringUtils.hpp"
#include "AttendeeIndex.hpp"
#include "PruneProblem.hpp"
#include <random>
#include <cmath>

using std::string;
using std::vector;
using std::pair;

typedef Xoshiro256StarStar Gen;

void usage()
{
    printf("Usage: anneal [<options>] <id> [<id> ...]\n");
    printf("  <id>  Problem to work on\n");
    printf("Options:\n");
    printf("  -h    Print usage information and exit\n");
    printf("  -d    Don't save changes to best solutions\n");
    printf("  -g    Use debug graphics\n");
    printf("  -r <seed1> <seed2>\n");
    printf("        Random number generator seeds (default: time & process id)\n");
    printf("  -s <s>\n");
    printf("        Limit each run to <s> seconds\n");
    printf("  -m <m>\n");
    printf("        Limit each run to <m> minutes\n");
    printf("  -h <h>\n");
    printf("        Limit each run to <h> hours\n");
    printf("  -i <i>\n");
    printf("        Limit each run to <i> iterations\n");
    printf("  -S <p>\n");
    printf("        Simplify problem by keeping <p> percent of closest attendees\n");
}

void simplifyProblem(Problem& problem, uint64_t simplifyPercent)
{
    double stageLowX = problem.stageX;
    double stageHighX = problem.stageX + problem.stageWidth;
    double stageLowY = problem.stageY;
    double stageHighY = problem.stageY + problem.stageHeight;

    auto calcStageD2 = [&](const Coord& coord) -> double
    {
        double dx = 0.0;
        if (coord.x < stageLowX) dx = stageLowX - coord.x;
        if (coord.x > stageHighX) dx = coord.x - stageHighX;
        double dy = 0.0;
        if (coord.y < stageLowY) dy = stageLowY - coord.y;
        if (coord.y > stageHighY) dy = coord.y - stageHighY;
        double d2 = dx * dx + dy * dy;
        return d2;
    };

    uint64_t attendeeCount = problem.attendees.size();
    vector<pair<double, uint64_t>> sorted(attendeeCount);

    for (uint64_t attendeeIdx = 0; attendeeIdx < attendeeCount; attendeeIdx++)
    {
        auto& attendee = problem.attendees[attendeeIdx];
        Coord attendeeCoord = attendee.coord;
        double d2 = calcStageD2(attendeeCoord);
        sorted[attendeeIdx].first = d2;
        sorted[attendeeIdx].second = attendeeIdx;
    }

    std::sort(sorted.begin(), sorted.end());

    uint64_t keepCount = attendeeCount * simplifyPercent / 100;

    vector<Attendee> newAttendees(keepCount);
    for (uint64_t keepIdx = 0; keepIdx < keepCount; keepIdx++)
    {
        newAttendees[keepIdx] = problem.attendees[sorted[keepIdx].second];
    }

    double keepD2 = sorted[keepCount - 1].first;

    vector<Pillar> newPillars;
    for (auto& pillar : problem.pillars)
    {
        double pillarD2 = calcStageD2(pillar.coord);
        if (pillarD2 <= keepD2)
        {
            newPillars.push_back(pillar);
        }
    }

    problem.attendees = std::move(newAttendees);
    problem.pillars = std::move(newPillars);
}

int main(int argc, char *argv[])
{
    bool gotLimit = false;
    bool gotSeeds = false;

    bool help = false;
    bool saveChanges = true;
    bool useDebugGraphics = false;
    bool simplify = false;
    uint64_t timeLimitMS = 60 * 60 * 1000;
    uint64_t iterLimit = 0;
    vector<uint64_t> ids;
    uint32_t seed1 = 1;
    uint32_t seed2 = 1;
    uint64_t simplifyPercent = 0;

    int iArg = 1;
    while (iArg < argc)
    {
        string strArg = argv[iArg++];

        if (strArg == "-h" || strArg == "--help")
        {
            help = true;
        }
        else if (strArg == "-d")
        {
            saveChanges = false;
        }
        else if (strArg == "-g")
        {
            useDebugGraphics = true;
        }
        else if (strArg == "-r")
        {
            if (gotSeeds || iArg >= argc)
            {
                usage();
                return 1;
            }
            strArg = argv[iArg++];
            if (!parseU32(strArg, &seed1))
            {
                usage();
                return 1;
            }
            if (iArg >= argc)
            {
                usage();
                return 1;
            }
            strArg = argv[iArg++];
            if (!parseU32(strArg, &seed2))
            {
                usage();
                return 1;
            }
            gotSeeds = true;
        }
        else if (strArg == "-s")
        {
            if (gotLimit || iArg >= argc)
            {
                usage();
                return 1;
            }
            strArg = argv[iArg++];
            if (!parseU64(strArg, &timeLimitMS) || timeLimitMS < 1 || timeLimitMS > 1000 * 60 * 60)
            {
                usage();
                return 1;
            }
            timeLimitMS *= 1000;
            iterLimit = 0;
            gotLimit = true;
        }
        else if (strArg == "-m")
        {
            if (gotLimit || iArg >= argc)
            {
                usage();
                return 1;
            }
            strArg = argv[iArg++];
            if (!parseU64(strArg, &timeLimitMS) || timeLimitMS < 1 || timeLimitMS > 1000 * 60)
            {
                usage();
                return 1;
            }
            timeLimitMS *= 1000 * 60;
            iterLimit = 0;
            gotLimit = true;
        }
        else if (strArg == "-h")
        {
            if (gotLimit || iArg >= argc)
            {
                usage();
                return 1;
            }
            strArg = argv[iArg++];
            if (!parseU64(strArg, &timeLimitMS) || timeLimitMS < 1 || timeLimitMS > 1000)
            {
                usage();
                return 1;
            }
            timeLimitMS *= 1000 * 60 * 60;
            iterLimit = 0;
            gotLimit = true;
        }
        else if (strArg == "-i")
        {
            if (gotLimit || iArg >= argc)
            {
                usage();
                return 1;
            }
            strArg = argv[iArg++];
            if (!parseU64(strArg, &iterLimit) || iterLimit < 1 || iterLimit > 1000000000)
            {
                usage();
                return 1;
            }
            timeLimitMS = 0;
            gotLimit = true;
        }
        else if (strArg == "-S")
        {
            if (simplify || iArg >= argc)
            {
                usage();
                return 1;
            }
            strArg = argv[iArg++];
            if (!parseU64(strArg, &simplifyPercent) || simplifyPercent < 1 || simplifyPercent > 99)
            {
                usage();
                return 1;
            }
            simplify = true;
        }
        else
        {
            uint64_t id = 0;
            if (!parseU64(strArg, &id) || id < 1 || id > problemCount)
            {
                usage();
                return 1;
            }
            ids.push_back(id);
        }
    }

    if (help)
    {
        usage();
        return 0;
    }

    if (ids.empty())
    {
        usage();
        return 1;
    }

    if (!gotSeeds)
    {
        uint64_t timeMS = getTimeMS();
        uint32_t pid = getProcessId();
        seed1 = (uint32_t)timeMS;
        seed2 = (uint32_t)pid ^ (uint32_t)(timeMS >> 32);
    }

    printf("seed1 = %" PRIu32 ", seed2 = %" PRIu32 "\n", seed1, seed2);

    uint64_t idIdx = 0;
    uint64_t idCount = ids.size();

    DebugGraphics debugGraphics;
    string text;
    if (useDebugGraphics)
    {
        debugGraphics.create();
        debugGraphics.update(Problem(), Solution(), text);
    }

    while (true)
    {
        uint64_t id = ids[idIdx];

        string msg;
        Problem realProblem;
        if (!parseProblem(id, &realProblem, &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }

        Problem problem(realProblem);
        if (simplify)
        {
            simplifyProblem(problem, simplifyPercent);
        }
        bool pruned = pruneProblem(problem);

        uint64_t musicianCount = problem.musicians.size();

        double stageLowX = problem.stageX + 10.0;
        double stageHighX = problem.stageX + problem.stageWidth - 10.0;
        double stageLowY = problem.stageY + 10.0;
        double stageHighY = problem.stageY + problem.stageHeight - 10.0;

        std::uniform_real_distribution<double> distX(stageLowX, stageHighX);
        std::uniform_real_distribution<double> distY(stageLowY, stageHighY);

        Best best;
        best.load(id);

        if (useDebugGraphics)
        {
            text = strprintf("Problem %" PRIu64 "", id);
            debugGraphics.update(problem, Solution(), text);
        }

        int64_t bestScore = 0;
        uint64_t bestCount = 0;

        Solution solution;
        solution.placements.resize(musicianCount);

        Solution curSolution;
        curSolution.placements.resize(musicianCount);

        AttendeeIndex attendeeIndex(problem);

        FastScore fastScore(attendeeIndex);

        while (true)
        {
            std::seed_seq seq{seed1, seed2};
            Gen gen(seq);

            for (uint64_t i = 0; i < musicianCount; i++)
            {
                while (true)
                {
                    bool ok = true;
                    Coord coord(distX(gen), distY(gen));
                    for (uint64_t j = 0; j < i; j++)
                    {
                        double d2 = calcD2(coord, solution.placements[j]);
                        if (d2 < 100.0)
                        {
                            ok = false;
                            break;
                        }
                    }
                    if (ok)
                    {
                        solution.placements[i] = coord;
                        break;
                    }
                }
            }

            if (!isValid(problem, solution))
            {
                printf("Not valid!\n");
                continue;
            }

            fastScore.init(problem, solution);
            int64_t score = fastScore.setVolumes(problem, solution);

#if 0
            if (score != calcScore(problem, solution))
            {
                printf("Score mismatch!\n");
                return 1;
            }
#endif

            curSolution = solution;
            int64_t curScore = score;

            if (useDebugGraphics)
            {
                debugGraphics.update(curSolution, text);
            }

            double allowedLossInitial = 1000000.0;
            double allowedLossFinal = 1.0;
            double lnRatio = std::log(allowedLossFinal / allowedLossInitial);

            double lnMinOffset = std::log(0.01);
            double lnMaxOffset = std::log(std::sqrt(calcL2(Coord(problem.stageWidth, problem.stageHeight))));
            std::uniform_real_distribution<double> lnOffsetDist(lnMinOffset, lnMaxOffset);
            std::uniform_real_distribution<double> angleDist(0.0, M_PI * 2);

            uint64_t passCount = 40;

            uint64_t startTimeMS = getTimeMS();

            uint64_t iter = 0;

            for (uint64_t pass = 0; pass < passCount; pass++)
            {
                double param = (double)pass / (double)passCount;
                double allowedLoss = allowedLossInitial * std::exp(lnRatio * param);

                uint64_t limitTimeMS = startTimeMS + timeLimitMS * (pass + 1) / passCount;
                uint64_t limitIter = iterLimit * (pass + 1) / passCount;

                while ((timeLimitMS != 0 && getTimeMS() < limitTimeMS) ||
                       (iterLimit != 0 && iter < limitIter))
                {
                    iter++;
                    uint64_t instrumentCount = problem.instruments.size();
                    bool isSwap = instrumentCount > 1 && fastRandom(10, gen) == 0;
                    if (isSwap)
                    {
                        uint64_t instrumentI = fastRandom(instrumentCount, gen);
                        uint64_t instrumentJ = fastRandom(instrumentCount - 1, gen);
                        if (instrumentJ >= instrumentI) instrumentJ++;
                        uint64_t musicianCountI = problem.instruments[instrumentI].size();
                        uint64_t musicianCountJ = problem.instruments[instrumentJ].size();
                        uint64_t i = problem.instruments[instrumentI][fastRandom(musicianCountI, gen)];
                        uint64_t j = problem.instruments[instrumentJ][fastRandom(musicianCountJ, gen)];
                        Coord oldCoordI = curSolution.placements[i];
                        Coord oldCoordJ = curSolution.placements[j];

                        fastScore.update(problem, curSolution, i, oldCoordJ);
                        fastScore.update(problem, curSolution, j, oldCoordI);
                        int64_t keepScore = fastScore.setVolumes(problem, curSolution);
#if 0
                        if (keepScore != calcScore(problem, curSolution))
                        {
                            printf("Score mismatch (swap)!\n");
                            return 1;
                        }
#endif
                        int64_t delta = keepScore - curScore;

                        bool keep = true;
                        if (delta < 0)
                        {
                            keep = fastRandomReal<double>(gen) * (double)(-delta) < allowedLoss;
                        }

                        if (keep)
                        {
                            curScore = keepScore;
                            fastScore.commit();
                        }
                        else
                        {
                            curSolution.placements[j] = oldCoordJ;
                            curSolution.placements[i] = oldCoordI;
                            fastScore.rollback();
                            fastScore.setVolumes(problem, curSolution);
                        }

                        if (keep && useDebugGraphics)
                        {
                            debugGraphics.update(curSolution, text);
                        }
                    }
                    else
                    {
                        uint64_t i = fastRandom(musicianCount, gen);
                        Coord oldCoord = curSolution.placements[i];
                        double lnOffset = lnOffsetDist(gen);
                        double offset = std::exp(lnOffset);
                        double angle = angleDist(gen);
                        double offsetX = offset * std::cos(angle);
                        double offsetY = offset * std::sin(angle);
                        Coord newCoord(oldCoord.x + offsetX, oldCoord.y + offsetY);
#if 0
                        if (newCoord.x < stageLowX ||
                            newCoord.x > stageHighX ||
                            newCoord.y < stageLowY ||
                            newCoord.y > stageHighY)
                        {
                            continue;
                        }
#endif
#if 0
                        if (newCoord.x < stageLowX)
                        {
                            newCoord.x = oldCoord.x - (oldCoord.x - stageLowX) * 0.5;
                        }
                        if (newCoord.x > stageHighX)
                        {
                            newCoord.x = oldCoord.x + (stageHighX - oldCoord.x) * 0.5;
                        }
                        if (newCoord.y < stageLowY)
                        {
                            newCoord.y = oldCoord.y - (oldCoord.y - stageLowY) * 0.5;
                        }
                        if (newCoord.y > stageHighY)
                        {
                            newCoord.y = oldCoord.y + (stageHighY - oldCoord.y) * 0.5;
                        }
#endif
#if 1
                        if (newCoord.x < stageLowX ||
                            newCoord.x > stageHighX)
                        {
                            newCoord.x = distX(gen);
                        }
                        if (newCoord.y < stageLowY ||
                            newCoord.y > stageHighY)
                        {
                            newCoord.y = distY(gen);
                        }
#endif
                        bool ok = true;
                        for (uint64_t j = 0; j < musicianCount; j++)
                        {
                            if (j == i) continue;
                            double d2 = calcD2(newCoord, curSolution.placements[j]);
                            if (d2 < 100.0)
                            {
                                ok = false;
                                break;
                            }
                        }
                        if (!ok) continue;

                        fastScore.update(problem, curSolution, i, newCoord);
                        int64_t keepScore = fastScore.setVolumes(problem, curSolution);
#if 0
                        if (keepScore != calcScore(problem, curSolution))
                        {
                            printf("Score mismatch (offset)!\n");
                            return 1;
                        }
#endif
                        int64_t delta = keepScore - curScore;

                        bool keep = true;
                        if (delta < 0)
                        {
                            keep = fastRandomReal<double>(gen) * (double)(-delta) < allowedLoss;
                        }

                        if (keep)
                        {
                            curScore = keepScore;
                            fastScore.commit();
                        }
                        else
                        {
                            curSolution.placements[i] = oldCoord;
                            fastScore.rollback();
                            fastScore.setVolumes(problem, curSolution);
                        }

                        if (keep && useDebugGraphics)
                        {
                            debugGraphics.update(curSolution, text);
                        }
                    }

                    if (curScore > score)
                    {
                        score = curScore;
                        solution = curSolution;
                        //printf("%2" PRIu64 "  pass %2" PRIu64 "/%2" PRIu64 "  iter %9" PRIu64 "  score %15s  curScore %15s\n", id, pass + 1, passCount, iter, formatWithCommas(score).c_str(), formatWithCommas(curScore).c_str());
                    }

                    if (useDebugGraphics && debugGraphics.destroyed()) break;
                }
#if 0
                if (curScore != calcScore(problem, curSolution))
                {
                    printf("Score mismatch!\n");
                    return 1;
                }
#endif
                printf("%2" PRIu64 "  pass %2" PRIu64 "/%2" PRIu64 "  iter %9" PRIu64 "  score %15s  curScore %15s\n", id, pass + 1, passCount, iter, formatWithCommas(score).c_str(), formatWithCommas(curScore).c_str());

                if (useDebugGraphics && debugGraphics.destroyed()) break;
                //break;
            }

            if (useDebugGraphics && debugGraphics.destroyed()) break;

#if 1
            if (score != calcScore(problem, solution))
            {
                printf("Score mismatch!\n");
                return 1;
            }
#endif

            bool better = (score > bestScore);
            bool same = (score == bestScore);
            bool newBest = (better || simplify) && saveChanges && best.update(solution);

            if (better)
            {
                bestScore = score;
                bestCount = 1;
            }
            else if (same)
            {
                bestCount++;
            }

            if (better || same)
            {
                int64_t realScore = score;
                if (simplify || pruned)
                {
                    realScore = calcScore(realProblem, solution);
                }

                printf("%2" PRIu64 "  %10" PRIu32 "  %10" PRIu32 "  %7" PRIu64 "  %15s  %s\n",
                       id,
                       seed1,
                       seed2,
                       bestCount,
                       formatWithCommas(realScore).c_str(),
                       newBest ? "  BEST" :
                       better ? "  best" :
                       same ? "  same" : "");
            }
            //break;

            seed2++;
            if (seed2 == 0)
            {
                seed2 = 1;
                seed1++;
                if (seed1 == 0)
                {
                    break;
                }
            }

            if (ids.size() > 1) break;
        }
        //break;

        if (useDebugGraphics && debugGraphics.destroyed()) break;

        if (seed1 == 0 && seed2 == 0) break;
        idIdx++;
        if (idIdx == idCount) idIdx = 0;
    }

    if (useDebugGraphics)
    {
        debugGraphics.destroy();
    }

    return 0;
}
