#include "Common.hpp"
#include "ProblemCount.hpp"
#include "ParseUtils.hpp"
#include "Problem.hpp"
#include "IsValid.hpp"
#include "CalcScore.hpp"
#include "Best.hpp"
#include "Xoshiro.hpp"
#include "TimeUtils.hpp"
#include "ProcessId.hpp"
#include "FormatUtils.hpp"
#include <random>

using std::string;

typedef Xoshiro256StarStar Gen;

void usage()
{
    printf("Usage: random [<options>] <id> [<seed1> <seed2>]\n");
    printf("  <id>  Problem to work on\n");
    printf("  <seed1>, <seed2>\n");
    printf("        Random number generator seeds (default: time & process id)\n");
    printf("Options:\n");
    printf("  -h    Print usage information and exit\n");
    printf("  -d    Don't save changes to best solutions\n");
}

int main(int argc, char *argv[])
{
    bool gotId = false;
    bool gotSeed1 = false;
    bool gotSeed2 = false;

    bool help = false;
    bool saveChanges = true;
    uint64_t id = 0;
    uint32_t seed1 = 1;
    uint32_t seed2 = 1;

    int iArg = 1;
    while (iArg < argc)
    {
        string strArg = argv[iArg++];

        if (strArg == "-h" || strArg == "--help")
        {
            help = true;
        }
        else if (strArg == "-d")
        {
            saveChanges = false;
        }
        else if (!gotId)
        {
            if (!parseU64(strArg, &id) || id < 1 || id > problemCount)
            {
                usage();
                return 1;
            }
            gotId = true;
        }
        else if (!gotSeed1 || !gotSeed2)
        {
            uint32_t seed = 0;
            if (!parseU32(strArg, &seed))
            {
                usage();
                return 1;
            }
            if (!gotSeed1)
            {
                seed1 = seed;
                gotSeed1 = true;
            }
            else
            {
                seed2 = seed;
                gotSeed2 = true;
            }
        }
        else
        {
            usage();
            return 1;
        }
    }

    if (help)
    {
        usage();
        return 0;
    }

    if (!gotId)
    {
        usage();
        return 1;
    }

    if (!gotSeed2)
    {
        if (!gotSeed1)
        {
            uint64_t timeMS = getTimeMS();
            uint32_t pid = getProcessId();
            seed1 = (uint32_t)timeMS;
            seed2 = (uint32_t)pid ^ (uint32_t)(timeMS >> 32);
        }
        else
        {
            usage();
            return 1;
        }
    }

    printf("seed1 = %" PRIu32 ", seed2 = %" PRIu32 "\n", seed1, seed2);

    string msg;
    Problem problem;
    if (!parseProblem(id, &problem, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }

    uint64_t musicianCount = problem.musicians.size();

    double stageLowX = problem.stageX + 10.0;
    double stageHighX = problem.stageX + problem.stageWidth - 10.0;
    double stageLowY = problem.stageY + 10.0;
    double stageHighY = problem.stageY + problem.stageHeight - 10.0;

    std::uniform_real_distribution<double> distX(stageLowX, stageHighX);
    std::uniform_real_distribution<double> distY(stageLowY, stageHighY);

    Best best;
    best.load(id);

    int64_t bestScore = 0;
    uint64_t bestCount = 0;

    Solution solution;

    solution.placements.resize(musicianCount);

    while (true)
    {
        std::seed_seq seq{seed1, seed2};
        Gen gen(seq);

        for (uint64_t i = 0; i < musicianCount; i++)
        {
            while (true)
            {
                bool ok = true;
                Coord coord(distX(gen), distY(gen));
                for (uint64_t j = 0; j < i; j++)
                {
                    double d2 = calcD2(coord, solution.placements[j]);
                    if (d2 < 100.0)
                    {
                        ok = false;
                        break;
                    }
                }
                if (ok)
                {
                    solution.placements[i] = coord;
                    break;
                }
            }
        }

        if (!isValid(problem, solution))
        {
            printf("Not valid!\n");
            continue;
        }

        int64_t score = calcScore(problem, solution);

        bool better = (score > bestScore);
        bool same = (score == bestScore);
        bool newBest = better && saveChanges && best.update(solution);

        if (better)
        {
            bestScore = score;
            bestCount = 1;
        }
        else if (same)
        {
            bestCount++;
        }

        if (better || same)
        {
            printf("%10" PRIu32 "  %10" PRIu32 "  %7" PRIu64 "  %15s  %s\n",
                   seed1,
                   seed2,
                   bestCount,
                   formatWithCommas(score).c_str(),
                   newBest ? "  BEST" :
                   better ? "  best" :
                   same ? "  same" : "");
        }

        seed2++;
        if (seed2 == 0)
        {
            seed2 = 1;
            seed1++;
            if (seed1 == 0)
            {
                break;
            }
        }
    }

    return 0;
}
