#include "Common.hpp"
#include "ParseUtils.hpp"
#include "Graphics.hpp"
#include "ProblemCount.hpp"
#include "Problem.hpp"
#include "DrawProblem.hpp"
#include "StringUtils.hpp"
#include "PruneProblem.hpp"

using std::string;

void usage()
{
    printf("Usage: show [<options>] [<id>]\n");
    printf("  <id>  Problem to show (default: 1)\n");
    printf("Options:\n");
    printf("  -h    Print usage information and exit\n");
}

int main(int argc, char *argv[])
{
    bool gotId = false;

    bool help = false;
    uint64_t id = 1;

    int iArg = 1;
    while (iArg < argc)
    {
        string strArg = argv[iArg++];

        if (strArg == "-h" || strArg == "--help")
        {
            help = true;
        }
        else if (!gotId)
        {
            if (!parseU64(strArg, &id) || id < 1 || id > problemCount)
            {
                usage();
                return 1;
            }
            gotId = true;
        }
        else
        {
            usage();
            return 1;
        }
    }

    if (help)
    {
        usage();
        return 0;
    }

    bool destroyRequested = false;
    bool destroyed = false;
    Graphics gr;
    bool prune = false;

    auto onDestroy = [&]()
    {
        destroyed = true;
    };
    gr.setOnDestroy(onDestroy);

    auto onPaint = [&]()
    {
        gr.beginPaint(true);

        Problem problem;
        Solution solution;
        string text;

        string msg;
        if (!parseProblem(id, &problem, &msg))
        {
            text = strprintf("Problem %" PRIu64 ": %s", id, msg.c_str());
        }
        else
        {
            if (prune)
            {
                pruneProblem(problem);
            }
            text = strprintf("Problem %" PRIu64 "", id);
            parseSolution(id, &solution);
        }

        drawProblem(gr, problem, solution, text);

        gr.endPaint();
    };
    gr.setOnPaint(onPaint);

    auto onKey = [&](Key key)
    {
        switch (key)
        {
        case Key::Esc:
        case Key::Q:
        case Key::ShiftQ:
        {
            if (!destroyRequested)
            {
                destroyRequested = true;
                gr.destroy();
            }
            break;
        }
        case Key::Right:
        {
            if (id < problemCount)
            {
                id++;
                gr.requestPaint();
            }
            break;
        }
        case Key::Left:
        {
            if (id > 1)
            {
                id--;
                gr.requestPaint();
            }
            break;
        }
        case Key::CtrlRight:
        {
            if (id < problemCount)
            {
                id += std::min((uint64_t)10, problemCount - id);
                gr.requestPaint();
            }
            break;
        }
        case Key::CtrlLeft:
        {
            if (id > 2)
            {
                id -= std::min((uint64_t)10, id - 1);
                gr.requestPaint();
            }
            break;
        }
        case Key::P:
        case Key::ShiftP:
        {
            prune = !prune;
            gr.requestPaint();
            break;
        }
        default: ;
        }
    };
    gr.setOnKey(onKey);

    string msg;
    if (!gr.create(1000, 1000, "Show", &msg))
    {
        fprintf(stderr, "%s\n", msg.c_str());
        return 1;
    }

    while (!destroyed)
    {
        if (Graphics::update(true))
        {
            continue;
        }
    }

    return 0;
}
