#include "Common.hpp"
#include "ParseUtils.hpp"
#include "Problem.hpp"
#include "ProblemCount.hpp"
#include "Cleanup.hpp"
#include "Best.hpp"
#include "Protocol.hpp"
#include "Scores.hpp"
#include "TimeUtils.hpp"
#include "FormatUtils.hpp"

using std::string;
using std::vector;

void usage()
{
    printf("Usage: submit [<options>] [<id>]\n");
    printf("  <n>   Problem id (default: all)\n");
    printf("Options:\n");
    printf("  -h    Print usage information and exit\n");
    printf("  -d    Dry run: don't actually submit\n");
}

int main(int argc, char *argv[])
{
    bool gotId = false;

    bool help = false;
    bool dryRun = false;
    uint64_t requestedId = 0;

    int iArg = 1;
    while (iArg < argc)
    {
        string strArg = argv[iArg++];

        if (strArg == "-h" || strArg == "--help")
        {
            help = true;
        }
        else if (strArg == "-d")
        {
            dryRun = true;
        }
        else if (!gotId)
        {
            if (!parseU64(strArg, &requestedId) || requestedId < 1 || requestedId > problemCount)
            {
                usage();
                return 1;
            }
            gotId = true;
        }
        else
        {
            usage();
            return 1;
        }
    }

    if (help)
    {
        usage();
        return 0;
    }

    string msg;

    Protocol::init();
    Cleanup cleanupProtocol([](){ Protocol::cleanup(); });

    string userboardData;
    if (!Protocol::getUserboard(&userboardData, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }

    Userboard userboard;
    if (!parseUserboard(userboardData, &userboard, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }

    if (!writeUserboard(userboardData))
    {
        printf("Error writing userboard\n");
        return 1;
    }

    vector<int64_t> myScores;
    getMyScores(userboard, &myScores);

    bool submittedAny = false;

    for (uint64_t id = 1; id <= problemCount; id++)
    {
        if (gotId && id != requestedId) continue;

        printf("%2" PRIu64 "  ", id);

        Problem problem;
        if (!parseProblem(id, &problem, &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }

        Best best;
        best.load(id);

        int64_t score = best.getScore();

        printf("%15s  ", formatWithCommas(score).c_str());

        int64_t submittedScore = myScores[id];

        if (score < submittedScore)
        {
            printf("WORSE than %15s!\n", formatWithCommas(submittedScore).c_str());
            continue;
        }
        else if (score == submittedScore)
        {
            printf("same\n");
            continue;
        }
        fflush(stdout);
        
        string solutionData;
        formatSolution(best.getSolution(), &solutionData);

        string submissionId;
        if (dryRun)
        {
            printf("would submit (dry run)\n");
        }
        else
        {
            if (!Protocol::submit(id, solutionData, &submissionId, &msg))
            {
                printf("ERROR!\n");
                printf("%s\n", msg.c_str());
                return 0;
            }

            submittedAny = true;
            printf("submitted (submission id = \"%s\")\n", submissionId.c_str());
        }
    }

    if (submittedAny)
    {
        sleepS(2);

        string userboardData;
        if (!Protocol::getUserboard(&userboardData, &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }

        Userboard userboard;
        if (!parseUserboard(userboardData, &userboard, &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }

        if (!writeUserboard(userboardData))
        {
            printf("Error writing userboard\n");
            return 1;
        }
    }

    return 0;
}
